@@ -10,12 +10,13 @@
 #include "common/buf.h"
 #include "common/font.h"
 #include "config/libinput.h"
+#include "resize_indicator.h"
 #include "theme.h"
 
 enum window_switcher_field_content {
 	LAB_FIELD_NONE = 0,
 	LAB_FIELD_TYPE,
-	LAB_FIELD_APP_ID,
+	LAB_FIELD_IDENTIFIER,
 	LAB_FIELD_TITLE,
 };
 
@@ -48,6 +49,7 @@ struct rcxml {
 	/* theme */
 	char *theme_name;
 	int corner_radius;
+	bool ssd_keep_border;
 	struct font font_activewindow;
 	struct font font_menuitem;
 	struct font font_osd;
@@ -77,6 +79,8 @@ struct rcxml {
 	int snap_edge_range;
 	bool snap_top_maximize;
 
+	enum resize_indicator_mode resize_indicator;
+
 	struct {
 		int popuptime;
 		int min_nr_workspaces;
