@@ -102,6 +102,8 @@ struct seat {
 	struct server *server;
 	struct wlr_keyboard_group *keyboard_group;
 
+	struct wl_list touch_points; /* struct touch_point.link */
+
 	/*
 	 * Enum of most recent server-side cursor image.  Set by
 	 * cursor_set().  Cleared when a client surface is entered
@@ -388,7 +390,6 @@ struct view *desktop_cycle_view(struct server *server, struct view *start_view,
 	enum lab_cycle_dir dir);
 struct view *desktop_focused_view(struct server *server);
 void desktop_focus_topmost_mapped_view(struct server *server);
-bool isfocusable(struct view *view);
 
 void keyboard_cancel_keybind_repeat(struct keyboard *keyboard);
 void keyboard_key_notify(struct wl_listener *listener, void *data);
