@@ -3,13 +3,14 @@
 #define LABWC_SSD_INTERNAL_H
 
 #include <wlr/util/box.h>
+#include "common/array-size.h"
 #include "ssd.h"
 
 #define FOR_EACH(tmp, ...) \
 { \
 	__typeof__(tmp) _x[] = { __VA_ARGS__, NULL }; \
 	size_t _i = 0; \
-	for ((tmp) = _x[_i]; _i < sizeof(_x) / sizeof(_x[0]) - 1; (tmp) = _x[++_i])
+	for ((tmp) = _x[_i]; _i < ARRAY_SIZE(_x) - 1; (tmp) = _x[++_i])
 
 #define FOR_EACH_END }
 
@@ -17,6 +18,7 @@ struct ssd_button {
 	struct view *view;
 	enum ssd_part_type type;
 	struct wlr_scene_node *hover;
+	struct wlr_scene_node *background;
 
 	struct wl_listener destroy;
 };
@@ -41,6 +43,7 @@ struct ssd {
 	 * don't update things we don't have to.
 	 */
 	struct {
+		bool squared_corners;   /* To un-round corner buttons on maximize */
 		struct wlr_box geometry;
 		struct ssd_state_title {
 			char *text;
@@ -54,14 +57,15 @@ struct ssd {
 
 	/* The top of the view, containing buttons, title, .. */
 	struct {
-		/* struct wlr_scene_tree *tree;      unused for now */
+		int height;
+		struct wlr_scene_tree *tree;
 		struct ssd_sub_tree active;
 		struct ssd_sub_tree inactive;
 	} titlebar;
 
 	/* Borders allow resizing as well */
 	struct {
-		/* struct wlr_scene_tree *tree;      unused for now */
+		struct wlr_scene_tree *tree;
 		struct ssd_sub_tree active;
 		struct ssd_sub_tree inactive;
 	} border;
