@@ -159,7 +159,7 @@ action_arg_add_int(struct action *action, const char *key, int value)
 }
 
 void
-action_arg_from_xml_node(struct action *action, char *nodename, char *content)
+action_arg_from_xml_node(struct action *action, const char *nodename, const char *content)
 {
 	assert(action);
 
@@ -181,7 +181,14 @@ action_arg_from_xml_node(struct action *action, char *nodename, char *content)
 	case ACTION_TYPE_MOVE_TO_EDGE:
 	case ACTION_TYPE_SNAP_TO_EDGE:
 		if (!strcmp(argument, "direction")) {
-			action_arg_add_str(action, argument, content);
+			enum view_edge edge = view_edge_parse(content);
+			if ((edge == VIEW_EDGE_CENTER && action->type != ACTION_TYPE_SNAP_TO_EDGE)
+					|| edge == VIEW_EDGE_INVALID) {
+				wlr_log(WLR_ERROR, "Invalid argument for action %s: '%s' (%s)",
+					action_names[action->type], argument, content);
+			} else {
+				action_arg_add_int(action, argument, edge);
+			}
 			goto cleanup;
 		}
 		break;
@@ -250,13 +257,13 @@ action_str_from_arg(struct action_arg *arg)
 }
 
 static bool
-arg_value_exists(struct action *action, const char *key)
+arg_value_exists(struct action *action, const char *key, enum action_arg_type type)
 {
 	assert(action);
 	assert(key);
 	struct action_arg *arg;
 	wl_list_for_each(arg, &action->args, link) {
-		if (!strcasecmp(key, arg->key)) {
+		if (!strcasecmp(key, arg->key) && arg->type == type) {
 			return true;
 		}
 	}
@@ -366,6 +373,8 @@ bool
 action_is_valid(struct action *action)
 {
 	const char *arg_name = NULL;
+	enum action_arg_type arg_type = LAB_ACTION_ARG_STR;
+
 	switch (action->type) {
 	case ACTION_TYPE_EXECUTE:
 		arg_name = "command";
@@ -373,6 +382,7 @@ action_is_valid(struct action *action)
 	case ACTION_TYPE_MOVE_TO_EDGE:
 	case ACTION_TYPE_SNAP_TO_EDGE:
 		arg_name = "direction";
+		arg_type = LAB_ACTION_ARG_INT;
 		break;
 	case ACTION_TYPE_SHOW_MENU:
 		arg_name = "menu";
@@ -392,7 +402,7 @@ action_is_valid(struct action *action)
 		return true;
 	}
 
-	if (arg_value_exists(action, arg_name)) {
+	if (arg_value_exists(action, arg_name, arg_type)) {
 		return true;
 	}
 
@@ -554,7 +564,7 @@ actions_run(struct view *activator, struct server *server,
 				struct buf cmd;
 				buf_init(&cmd);
 				buf_add(&cmd, action_str_from_arg(arg));
-				buf_expand_shell_variables(&cmd);
+				buf_expand_tilde(&cmd);
 				spawn_async_no_shell(cmd.buf);
 				free(cmd.buf);
 			}
@@ -564,13 +574,16 @@ actions_run(struct view *activator, struct server *server,
 			break;
 		case ACTION_TYPE_MOVE_TO_EDGE:
 			if (view) {
-				view_move_to_edge(view, action_str_from_arg(arg));
+				/* Config parsing makes sure that direction is a valid direction */
+				enum view_edge edge = get_arg_value_int(action, "direction", 0);
+				view_move_to_edge(view, edge);
 			}
 			break;
 		case ACTION_TYPE_SNAP_TO_EDGE:
 			if (view) {
-				view_snap_to_edge(view, action_str_from_arg(arg),
-					/*store_natural_geometry*/ true);
+				/* Config parsing makes sure that direction is a valid direction */
+				enum view_edge edge = get_arg_value_int(action, "direction", 0);
+				view_snap_to_edge(view, edge, /*store_natural_geometry*/ true);
 			}
 			break;
 		case ACTION_TYPE_NEXT_WINDOW:
@@ -667,11 +680,10 @@ actions_run(struct view *activator, struct server *server,
 			}
 			break;
 		case ACTION_TYPE_MOVE_RELATIVE:
-			if (view && !view->fullscreen) {
+			if (view) {
 				int x = get_arg_value_int(action, "x", 0);
 				int y = get_arg_value_int(action, "y", 0);
-				view_maximize(view, false, false);
-				view_move(view, view->pending.x + x, view->pending.y + y);
+				view_move_relative(view, x, y);
 			}
 			break;
 		case ACTION_TYPE_SEND_TO_DESKTOP:
