@@ -150,6 +150,8 @@ fill_window_rule(char *nodename, char *content)
 		set_property(content, &current_window_rule->skip_taskbar);
 	} else if (!strcasecmp(nodename, "skipWindowSwitcher")) {
 		set_property(content, &current_window_rule->skip_window_switcher);
+	} else if (!strcasecmp(nodename, "ignoreFocusRequest")) {
+		set_property(content, &current_window_rule->ignore_focus_request);
 
 	/* Actions */
 	} else if (!strcmp(nodename, "name.action")) {
@@ -183,8 +185,11 @@ fill_window_switcher_field(char *nodename, char *content)
 	} else if (!strcmp(nodename, "content")) {
 		if (!strcmp(content, "type")) {
 			current_field->content = LAB_FIELD_TYPE;
+		} else if (!strcmp(content, "identifier")) {
+			current_field->content = LAB_FIELD_IDENTIFIER;
 		} else if (!strcmp(content, "app_id")) {
-			current_field->content = LAB_FIELD_APP_ID;
+			wlr_log(WLR_ERROR, "window-switcher field 'app_id' is deprecated");
+			current_field->content = LAB_FIELD_IDENTIFIER;
 		} else if (!strcmp(content, "title")) {
 			current_field->content = LAB_FIELD_TITLE;
 		} else {
@@ -262,6 +267,8 @@ fill_keybind(char *nodename, char *content)
 	} else if (!current_keybind) {
 		wlr_log(WLR_ERROR, "expect <keybind key=\"\"> element first. "
 			"nodename: '%s' content: '%s'", nodename, content);
+	} else if (!strcasecmp(nodename, "layoutDependent")) {
+		set_bool(content, &current_keybind->use_syms_only);
 	} else if (!strcmp(nodename, "name.action")) {
 		current_keybind_action = action_create(content);
 		if (current_keybind_action) {
@@ -336,11 +343,11 @@ fill_mousebind(char *nodename, char *content)
 	}
 }
 
-static enum libinput_config_accel_profile
+static int
 get_accel_profile(const char *s)
 {
 	if (!s) {
-		return LIBINPUT_CONFIG_ACCEL_PROFILE_FLAT;
+		return -1;
 	}
 	if (!strcasecmp(s, "flat")) {
 		return LIBINPUT_CONFIG_ACCEL_PROFILE_FLAT;
@@ -348,7 +355,7 @@ get_accel_profile(const char *s)
 	if (!strcasecmp(s, "adaptive")) {
 		return LIBINPUT_CONFIG_ACCEL_PROFILE_ADAPTIVE;
 	}
-	return LIBINPUT_CONFIG_ACCEL_PROFILE_FLAT;
+	return -1;
 }
 
 static void
@@ -405,6 +412,22 @@ fill_libinput_category(char *nodename, char *content)
 		} else {
 			wlr_log(WLR_ERROR, "invalid tapButtonMap");
 		}
+	} else if (!strcasecmp(nodename, "tapAndDrag")) {
+		int ret = parse_bool(content, -1);
+		if (ret < 0) {
+			return;
+		}
+		current_libinput_category->tap_and_drag = ret
+			? LIBINPUT_CONFIG_DRAG_ENABLED
+			: LIBINPUT_CONFIG_DRAG_DISABLED;
+	} else if (!strcasecmp(nodename, "dragLock")) {
+		int ret = parse_bool(content, -1);
+		if (ret < 0) {
+			return;
+		}
+		current_libinput_category->drag_lock = ret
+			? LIBINPUT_CONFIG_DRAG_LOCK_ENABLED
+			: LIBINPUT_CONFIG_DRAG_LOCK_DISABLED;
 	} else if (!strcasecmp(nodename, "accelProfile")) {
 		current_libinput_category->accel_profile =
 			get_accel_profile(content);
@@ -575,6 +598,8 @@ entry(xmlNode *node, char *nodename, char *content)
 		rc.theme_name = xstrdup(content);
 	} else if (!strcmp(nodename, "cornerradius.theme")) {
 		rc.corner_radius = atoi(content);
+	} else if (!strcasecmp(nodename, "keepBorder.theme")) {
+		set_bool(content, &rc.ssd_keep_border);
 	} else if (!strcmp(nodename, "name.font.theme")) {
 		fill_font(nodename, content, font_place);
 	} else if (!strcmp(nodename, "size.font.theme")) {
@@ -655,6 +680,16 @@ entry(xmlNode *node, char *nodename, char *content)
 		rc.workspace_config.popuptime = atoi(content);
 	} else if (!strcasecmp(nodename, "number.desktops")) {
 		rc.workspace_config.min_nr_workspaces = MAX(1, atoi(content));
+	} else if (!strcasecmp(nodename, "popupShow.resize")) {
+		if (!strcasecmp(content, "Always")) {
+			rc.resize_indicator = LAB_RESIZE_INDICATOR_ALWAYS;
+		} else if (!strcasecmp(content, "Never")) {
+			rc.resize_indicator = LAB_RESIZE_INDICATOR_NEVER;
+		} else if (!strcasecmp(content, "Nonpixel")) {
+			rc.resize_indicator = LAB_RESIZE_INDICATOR_NON_PIXEL;
+		} else {
+			wlr_log(WLR_ERROR, "Invalid value for <resize popupShow />");
+		}
 	}
 }
 
@@ -780,6 +815,7 @@ rcxml_init(void)
 	has_run = true;
 
 	rc.xdg_shell_server_side_deco = true;
+	rc.ssd_keep_border = true;
 	rc.corner_radius = 8;
 
 	init_font_defaults(&rc.font_activewindow);
@@ -803,6 +839,8 @@ rcxml_init(void)
 	rc.window_switcher.preview = true;
 	rc.window_switcher.outlines = true;
 
+	rc.resize_indicator = LAB_RESIZE_INDICATOR_NEVER;
+
 	rc.workspace_config.popuptime = INT_MIN;
 	rc.workspace_config.min_nr_workspaces = 1;
 }
@@ -847,7 +885,8 @@ load_default_key_bindings(void)
 		wl_list_append(&k->actions, &action->link);
 
 		if (key_combos[i].attribute && key_combos[i].value) {
-			action_arg_add_str(action, key_combos[i].attribute, key_combos[i].value);
+			action_arg_from_xml_node(action,
+				key_combos[i].attribute, key_combos[i].value);
 		}
 	}
 }
@@ -956,7 +995,8 @@ load_default_mouse_bindings(void)
 		 * slightly more sophisticated approach will be needed.
 		 */
 		if (current->attribute && current->value) {
-			action_arg_add_str(action, current->attribute, current->value);
+			action_arg_from_xml_node(action,
+				current->attribute, current->value);
 		}
 	}
 	wlr_log(WLR_DEBUG, "Loaded %u merged mousebinds", count);
@@ -1037,7 +1077,7 @@ static struct {
 	int width;
 } fields[] = {
 	{ LAB_FIELD_TYPE, 25 },
-	{ LAB_FIELD_APP_ID, 25 },
+	{ LAB_FIELD_IDENTIFIER, 25 },
 	{ LAB_FIELD_TITLE, 50 },
 	{ LAB_FIELD_NONE, 0 },
 };
