@@ -7,6 +7,7 @@
 #include <wlr/types/wlr_primary_selection.h>
 #include <wlr/util/region.h>
 #include "action.h"
+#include "common/array-size.h"
 #include "common/mem.h"
 #include "common/scene-helpers.h"
 #include "config/mousebind.h"
@@ -51,14 +52,12 @@ static const char * const cursors_x11[] = {
 	"left_side"
 };
 
-#define ARRAY_SIZE(array) (sizeof(array) / sizeof((array)[0]))
 static_assert(
 	ARRAY_SIZE(cursors_xdg) == LAB_CURSOR_COUNT,
 	"XDG cursor names are out of sync");
 static_assert(
 	ARRAY_SIZE(cursors_x11) == LAB_CURSOR_COUNT,
 	"X11 cursor names are out of sync");
-#undef ARRAY_SIZE
 
 enum lab_cursors
 cursor_get_from_edge(uint32_t resize_edges)
@@ -300,8 +299,7 @@ update_pressed_surface(struct seat *seat, struct cursor_context *ctx)
 	 * context menus (in contrast) do not use an XDG popup grab and
 	 * do not work properly if we send leave/enter events.
 	 */
-	if (seat->seat->pointer_state.grab ==
-			seat->seat->pointer_state.default_grab) {
+	if (!wlr_seat_pointer_has_grab(seat->seat)) {
 		return false;
 	}
 	if (seat->pressed.surface && ctx->surface != seat->pressed.surface) {
@@ -922,6 +920,23 @@ cursor_button_press(struct seat *seat, struct wlr_pointer_button_event *event)
 			seat_set_focus_layer(seat, layer);
 		}
 	}
+	if (ctx.type == LAB_SSD_LAYER_SUBSURFACE) {
+		seat_focus_surface(seat, ctx.surface);
+	}
+
+	if (ctx.type != LAB_SSD_CLIENT && ctx.type != LAB_SSD_LAYER_SUBSURFACE
+			&& wlr_seat_pointer_has_grab(seat->seat)) {
+		/*
+		 * If we have an active popup grab (an open popup) we want to
+		 * cancel that grab whenever the user presses on anything that
+		 * is not the client itself, for example the desktop or any
+		 * part of the server side decoration.
+		 *
+		 * Note: This does not work for XWayland clients
+		 */
+		wlr_seat_pointer_end_grab(seat->seat);
+		return;
+	}
 
 	/* Bindings to the Frame context swallow mouse events if activated */
 	bool consumed_by_frame_context =
@@ -943,16 +958,6 @@ cursor_button_release(struct seat *seat, struct wlr_pointer_button_event *event)
 
 	seat_reset_pressed(seat);
 
-	if (pressed_surface && ctx.surface != pressed_surface) {
-		/*
-		 * Button released but originally pressed over a different surface.
-		 * Just send the release event to the still focused surface.
-		 */
-		wlr_seat_pointer_notify_button(seat->seat, event->time_msec,
-			event->button, event->state);
-		return;
-	}
-
 	if (server->input_mode == LAB_INPUT_STATE_MENU) {
 		if (close_menu) {
 			menu_close_root(server);
@@ -966,6 +971,22 @@ cursor_button_release(struct seat *seat, struct wlr_pointer_button_event *event)
 	if (server->input_mode != LAB_INPUT_STATE_PASSTHROUGH) {
 		/* Exit interactive move/resize mode */
 		interactive_finish(server->grabbed_view);
+
+		if (pressed_surface) {
+			/* Ensure CSD clients see the release event */
+			wlr_seat_pointer_notify_button(seat->seat, event->time_msec,
+				event->button, event->state);
+		}
+		return;
+	}
+
+	if (pressed_surface && ctx.surface != pressed_surface) {
+		/*
+		 * Button released but originally pressed over a different surface.
+		 * Just send the release event to the still focused surface.
+		 */
+		wlr_seat_pointer_notify_button(seat->seat, event->time_msec,
+			event->button, event->state);
 		return;
 	}
 
