@@ -37,6 +37,8 @@ static struct menu *current_menu;
 static struct menu *menus;
 static int nr_menus, alloc_menus;
 
+/* TODO: split this whole file into parser.c and actions.c*/
+
 static struct menu *
 menu_create(struct server *server, const char *id, const char *label)
 {
@@ -232,7 +234,7 @@ item_create(struct menu *menu, const char *text, bool show_arrow)
 	/* Update menu extents */
 	menu->size.height += menuitem->height;
 
-	wl_list_insert(&menu->menuitems, &menuitem->link);
+	wl_list_append(&menu->menuitems, &menuitem->link);
 	wl_list_init(&menuitem->actions);
 	return menuitem;
 }
@@ -272,7 +274,7 @@ separator_create(struct menu *menu, const char *label)
 		theme->menu_separator_padding_height);
 
 	menu->size.height += menuitem->height;
-	wl_list_insert(&menu->menuitems, &menuitem->link);
+	wl_list_append(&menu->menuitems, &menuitem->link);
 	wl_list_init(&menuitem->actions);
 	return menuitem;
 }
@@ -571,7 +573,7 @@ menu_get_full_width(struct menu *menu)
 	int child_width;
 	int max_child_width = 0;
 	struct menuitem *item;
-	wl_list_for_each_reverse(item, &menu->menuitems, link) {
+	wl_list_for_each(item, &menu->menuitems, link) {
 		if (!item->submenu) {
 			continue;
 		}
@@ -636,7 +638,7 @@ menu_configure(struct menu *menu, int lx, int ly, enum menu_align align)
 	int rel_y;
 	int new_lx, new_ly;
 	struct menuitem *item;
-	wl_list_for_each_reverse(item, &menu->menuitems, link) {
+	wl_list_for_each(item, &menu->menuitems, link) {
 		if (!item->submenu) {
 			continue;
 		}
@@ -663,7 +665,7 @@ menu_hide_submenu(const char *id)
 		menu = menus + i;
 		bool should_reposition = false;
 		struct menuitem *item, *next;
-		wl_list_for_each_reverse_safe(item, next, &menu->menuitems, link) {
+		wl_list_for_each_safe(item, next, &menu->menuitems, link) {
 			if (item->submenu == hide_menu) {
 				item_destroy(item);
 				should_reposition = true;
@@ -675,7 +677,7 @@ menu_hide_submenu(const char *id)
 		}
 		/* Re-position items vertically */
 		menu->size.height = 0;
-		wl_list_for_each_reverse(item, &menu->menuitems, link) {
+		wl_list_for_each(item, &menu->menuitems, link) {
 			wlr_scene_node_set_position(&item->tree->node, 0,
 				menu->size.height);
 			menu->size.height += item->height;
@@ -841,22 +843,14 @@ menu_open(struct menu *menu, int x, int y)
 	menu->server->input_mode = LAB_INPUT_STATE_MENU;
 }
 
-void
-menu_process_cursor_motion(struct wlr_scene_node *node)
+static void
+menu_process_item_selection(struct menuitem *item)
 {
-	assert(node && node->data);
-	struct menuitem *item = node_menuitem_from_node(node);
-
-	/* This function shall only be called for menu nodes */
 	assert(item);
 
 	if (!item->selectable) {
 		return;
 	}
-	if (node == &item->selected.tree->node) {
-		/* We are on an already selected item */
-		return;
-	}
 
 	/* We are on an item that has new mouse-focus */
 	menu_set_selection(item->parent, item);
@@ -868,6 +862,8 @@ menu_process_cursor_motion(struct wlr_scene_node *node)
 	if (item->submenu) {
 		/* Sync the triggering view */
 		item->submenu->triggered_by_view = item->parent->triggered_by_view;
+		/* Ensure the submenu has its parent set correctly */
+		item->submenu->parent = item->parent;
 		/* And open the new submenu tree */
 		wlr_scene_node_set_enabled(
 			&item->submenu->scene_tree->node, true);
@@ -875,14 +871,61 @@ menu_process_cursor_motion(struct wlr_scene_node *node)
 	item->parent->selection.menu = item->submenu;
 }
 
-bool
-menu_call_actions(struct wlr_scene_node *node)
+/* Get the deepest submenu with active item selection or the root menu itself */
+static struct menu *
+get_selection_leaf(struct server *server)
 {
-	assert(node && node->data);
-	struct menuitem *item = node_menuitem_from_node(node);
+	struct menu *menu = server->menu_current;
+	if (!menu) {
+		return NULL;
+	}
 
-	if (item->submenu) {
-		/* We received a click on an item that just opens a submenu */
+	while (menu->selection.menu) {
+		if (!menu->selection.menu->selection.item) {
+			return menu;
+		}
+		menu = menu->selection.menu;
+	}
+
+	return menu;
+}
+
+/* Selects the next or previous sibling of the currently selected item */
+static void
+menu_item_select(struct server *server, bool forward)
+{
+	struct menu *menu = get_selection_leaf(server);
+	if (!menu) {
+		return;
+	}
+
+	struct menuitem *item = NULL;
+	struct menuitem *selection = menu->selection.item;
+	struct wl_list *start = selection ? &selection->link : &menu->menuitems;
+	struct wl_list *current = start;
+	while (!item || !item->selectable) {
+		current = forward ? current->next : current->prev;
+		if (current == start) {
+			return;
+		}
+		if (current == &menu->menuitems) {
+			/* Allow wrap around */
+			item = NULL;
+			continue;
+		}
+		item = wl_container_of(current, item, link);
+	}
+
+	menu_process_item_selection(item);
+}
+
+static bool
+menu_execute_item(struct menuitem *item)
+{
+	assert(item);
+
+	if (item->submenu || !item->selectable) {
+		/* We received a click on a separator or item that just opens a submenu */
 		return false;
 	}
 
@@ -900,6 +943,89 @@ menu_call_actions(struct wlr_scene_node *node)
 	return true;
 }
 
+/* Keyboard based selection */
+void
+menu_item_select_next(struct server *server)
+{
+	menu_item_select(server, /* forward */ true);
+}
+
+void
+menu_item_select_previous(struct server *server)
+{
+	menu_item_select(server, /* forward */ false);
+}
+
+bool
+menu_call_selected_actions(struct server *server)
+{
+	struct menu *menu = get_selection_leaf(server);
+	if (!menu || !menu->selection.item) {
+		return false;
+	}
+
+	return menu_execute_item(menu->selection.item);
+}
+
+/* Selects the first item on the submenu attached to the current selection */
+void
+menu_submenu_enter(struct server *server)
+{
+	struct menu *menu = get_selection_leaf(server);
+	if (!menu || !menu->selection.menu) {
+		return;
+	}
+
+	struct wl_list *start = &menu->selection.menu->menuitems;
+	struct wl_list *current = start;
+	struct menuitem *item = NULL;
+	while (!item || !item->selectable) {
+		current = current->next;
+		if (current == start) {
+			return;
+		}
+		item = wl_container_of(current, item, link);
+	}
+
+	menu_process_item_selection(item);
+}
+
+/* Re-selects the selected item on the parent menu of the current selection */
+void
+menu_submenu_leave(struct server *server)
+{
+	struct menu *menu = get_selection_leaf(server);
+	if (!menu || !menu->parent || !menu->parent->selection.item) {
+		return;
+	}
+
+	menu_process_item_selection(menu->parent->selection.item);
+}
+
+/* Mouse based selection */
+void
+menu_process_cursor_motion(struct wlr_scene_node *node)
+{
+	assert(node && node->data);
+	struct menuitem *item = node_menuitem_from_node(node);
+
+	if (item->selectable && node == &item->selected.tree->node) {
+		/* We are on an already selected item */
+		return;
+	}
+
+	menu_process_item_selection(item);
+}
+
+bool
+menu_call_actions(struct wlr_scene_node *node)
+{
+	assert(node && node->data);
+	struct menuitem *item = node_menuitem_from_node(node);
+
+	return menu_execute_item(item);
+}
+
 void
 menu_close_root(struct server *server)
 {
