@@ -7,6 +7,7 @@
 #include <wlr/util/log.h>
 #include <wlr/util/box.h>
 #include "buffer.h"
+#include "common/array.h"
 #include "common/buf.h"
 #include "common/font.h"
 #include "common/graphic-helpers.h"
@@ -258,12 +259,11 @@ get_title(struct view *view)
 static void
 render_osd(struct server *server, cairo_t *cairo, int w, int h,
 		struct wl_list *node_list, bool show_workspace,
-		const char *workspace_name)
+		const char *workspace_name, struct wl_array *views)
 {
 	struct view *cycle_view = server->osd_state.cycle_view;
 	struct theme *theme = server->theme;
 
-	struct wlr_scene_node *node;
 	cairo_surface_t *surf = cairo_get_target(cairo);
 
 	/* Draw background */
@@ -317,17 +317,8 @@ render_osd(struct server *server, cairo_t *cairo, int w, int h,
 		- 2 * theme->osd_window_switcher_item_active_border_width;
 
 	/* Draw text for each node */
-	wl_list_for_each_reverse(node, node_list, link) {
-		if (!node->data) {
-			/* We found some non-view, most likely the region overlay */
-			continue;
-		}
-		struct view *view = node_view_from_node(node);
-		enum property skip = window_rules_get_property(view, "skipWindowSwitcher");
-		if (!isfocusable(view) || skip == LAB_PROP_TRUE) {
-			continue;
-		}
-
+	struct view **view;
+	wl_array_for_each(view, views) {
 		/*
 		 *    OSD border
 		 * +---------------------------------+
@@ -359,13 +350,13 @@ render_osd(struct server *server, cairo_t *cairo, int w, int h,
 
 			switch (field->content) {
 			case LAB_FIELD_TYPE:
-				buf_add(&buf, get_type(view));
+				buf_add(&buf, get_type(*view));
 				break;
-			case LAB_FIELD_APP_ID:
-				buf_add(&buf, get_app_id(view));
+			case LAB_FIELD_IDENTIFIER:
+				buf_add(&buf, get_app_id(*view));
 				break;
 			case LAB_FIELD_TITLE:
-				buf_add(&buf, get_title(view));
+				buf_add(&buf, get_title(*view));
 				break;
 			default:
 				break;
@@ -379,7 +370,7 @@ render_osd(struct server *server, cairo_t *cairo, int w, int h,
 			x += field_width + theme->osd_window_switcher_item_padding_x;
 		}
 
-		if (view == cycle_view) {
+		if (*view == cycle_view) {
 			/* Highlight current window */
 			struct wlr_fbox fbox = {
 				.x = theme->osd_border_width + theme->osd_window_switcher_padding,
@@ -402,33 +393,6 @@ render_osd(struct server *server, cairo_t *cairo, int w, int h,
 	cairo_surface_flush(surf);
 }
 
-static int
-get_osd_height(struct wl_list *node_list)
-{
-	int height = 0;
-	struct view *view;
-	struct wlr_scene_node *node;
-	wl_list_for_each(node, node_list, link) {
-		if (!node->data) {
-			/* We found some non-view, most likely the region overlay */
-			continue;
-		}
-		view = node_view_from_node(node);
-		enum property skip = window_rules_get_property(view, "skipWindowSwitcher");
-		if (!isfocusable(view) || skip == LAB_PROP_TRUE) {
-			continue;
-		}
-
-		/* Include item border width */
-		height += rc.theme->osd_window_switcher_item_height;
-	}
-
-	/* Add OSD border width */
-	height += 2 * rc.theme->osd_border_width;
-	height += 2 * rc.theme->osd_window_switcher_padding;
-	return height;
-}
-
 static void
 display_osd(struct output *output)
 {
@@ -439,9 +403,18 @@ display_osd(struct output *output)
 	bool show_workspace = wl_list_length(&rc.workspace_config.workspaces) > 1;
 	const char *workspace_name = server->workspace_current->name;
 
+	struct wl_array views;
+	wl_array_init(&views);
+	view_array_append(server, &views,
+		LAB_VIEW_CRITERIA_CURRENT_WORKSPACE
+		| LAB_VIEW_CRITERIA_NO_ALWAYS_ON_TOP
+		| LAB_VIEW_CRITERIA_NO_SKIP_WINDOW_SWITCHER);
+
 	float scale = output->wlr_output->scale;
 	int w = theme->osd_window_switcher_width;
-	int h = get_osd_height(node_list);
+	int h = wl_array_len(&views) * rc.theme->osd_window_switcher_item_height
+		+ 2 * rc.theme->osd_border_width
+		+ 2 * rc.theme->osd_window_switcher_padding;
 	if (show_workspace) {
 		/* workspace indicator */
 		h += theme->osd_window_switcher_item_height;
@@ -452,10 +425,16 @@ display_osd(struct output *output)
 		wlr_buffer_drop(&output->osd_buffer->base);
 	}
 	output->osd_buffer = buffer_create_cairo(w, h, scale, true);
+	if (!output->osd_buffer) {
+		wlr_log(WLR_ERROR, "Failed to allocate cairo buffer for the window switcher");
+		return;
+	}
 
 	/* Render OSD image */
 	cairo_t *cairo = output->osd_buffer->cairo;
-	render_osd(server, cairo, w, h, node_list, show_workspace, workspace_name);
+	render_osd(server, cairo, w, h, node_list, show_workspace,
+		workspace_name, &views);
+	wl_array_release(&views);
 
 	struct wlr_scene_buffer *scene_buffer = wlr_scene_buffer_create(
 		output->osd_tree, &output->osd_buffer->base);
@@ -487,7 +466,7 @@ osd_update(struct server *server)
 		return;
 	}
 
-	if (rc.window_switcher.show) {
+	if (rc.window_switcher.show && rc.theme->osd_window_switcher_width > 0) {
 		/* Display the actual OSD */
 		struct output *output;
 		wl_list_for_each(output, &server->outputs, link) {
@@ -500,7 +479,7 @@ osd_update(struct server *server)
 
 	/* Outline current window */
 	if (rc.window_switcher.outlines) {
-		if (isfocusable(server->osd_state.cycle_view)) {
+		if (view_isfocusable(server->osd_state.cycle_view)) {
 			osd_update_preview_outlines(server->osd_state.cycle_view);
 		}
 	}
