@@ -95,8 +95,26 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 			dc->tap_button_map);
 	}
 
+	if (libinput_device_config_tap_get_finger_count(libinput_dev) <= 0
+			|| dc->tap_and_drag < 0) {
+		wlr_log(WLR_INFO, "tap-and-drag not configured");
+	} else {
+		wlr_log(WLR_INFO, "tap-and-drag configured");
+		libinput_device_config_tap_set_drag_enabled(
+			libinput_dev, dc->tap_and_drag);
+	}
+
+	if (libinput_device_config_tap_get_finger_count(libinput_dev) <= 0
+			|| dc->drag_lock < 0) {
+		wlr_log(WLR_INFO, "drag lock not configured");
+	} else {
+		wlr_log(WLR_INFO, "drag lock configured");
+		libinput_device_config_tap_set_drag_lock_enabled(
+			libinput_dev, dc->drag_lock);
+	}
+
 	if (libinput_device_config_scroll_has_natural_scroll(libinput_dev) <= 0
-		|| dc->natural_scroll < 0) {
+			|| dc->natural_scroll < 0) {
 		wlr_log(WLR_INFO, "natural scroll not configured");
 	} else {
 		wlr_log(WLR_INFO, "natural scroll configured");
@@ -105,7 +123,7 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 	}
 
 	if (libinput_device_config_left_handed_is_available(libinput_dev) <= 0
-		|| dc->left_handed < 0) {
+			|| dc->left_handed < 0) {
 		wlr_log(WLR_INFO, "left-handed mode not configured");
 	} else {
 		wlr_log(WLR_INFO, "left-handed mode configured");
@@ -137,7 +155,7 @@ configure_libinput(struct wlr_input_device *wlr_input_device)
 	}
 
 	if (libinput_device_config_dwt_is_available(libinput_dev) == 0
-		|| dc->dwt < 0) {
+			|| dc->dwt < 0) {
 		wlr_log(WLR_INFO, "dwt not configured");
 	} else {
 		wlr_log(WLR_INFO, "dwt configured");
@@ -329,6 +347,7 @@ seat_init(struct server *server)
 		exit(EXIT_FAILURE);
 	}
 
+	wl_list_init(&seat->touch_points);
 	wl_list_init(&seat->constraint_commit.link);
 	wl_list_init(&seat->inputs);
 	seat->new_input.notify = new_input_notify;
