@@ -2,10 +2,11 @@
 /*
  * Theme engine for labwc
  *
- * Copyright (C) Johan Malm 2020-2021
+ * Copyright (C) Johan Malm 2020-2023
  */
 
 #define _POSIX_C_SOURCE 200809L
+#include "config.h"
 #include <cairo.h>
 #include <ctype.h>
 #include <drm_fourcc.h>
@@ -18,17 +19,135 @@
 #include <wlr/util/box.h>
 #include <wlr/util/log.h>
 #include <strings.h>
+#include "common/array-size.h"
 #include "common/dir.h"
 #include "common/font.h"
 #include "common/graphic-helpers.h"
 #include "common/match.h"
 #include "common/string-helpers.h"
 #include "config/rcxml.h"
+#include "button/button-png.h"
+
+#if HAVE_RSVG
+#include "button/button-svg.h"
+#endif
+
+#include "button/button-xbm.h"
 #include "theme.h"
-#include "xbm/xbm.h"
 #include "buffer.h"
 #include "ssd.h"
 
+struct button {
+	const char *name;
+	char fallback_button[6];	/* built-in 6x6 button */
+	struct {
+		struct lab_data_buffer **buffer;
+		float *rgba;
+	} active, inactive;
+};
+
+static void
+drop(struct lab_data_buffer **buffer)
+{
+	if (*buffer) {
+		wlr_buffer_drop(&(*buffer)->base);
+		*buffer = NULL;
+	}
+}
+
+static void
+load_buttons(struct theme *theme)
+{
+	struct button buttons[] = {
+		{
+			"menu",
+			{ 0x00, 0x18, 0x3c, 0x3c, 0x18, 0x00 },
+			{
+				&theme->button_menu_active_unpressed,
+				theme->window_active_button_menu_unpressed_image_color,
+			},
+			{
+				&theme->button_menu_inactive_unpressed,
+				theme->window_inactive_button_menu_unpressed_image_color,
+			},
+		},
+		{
+			"iconify",
+			{ 0x00, 0x00, 0x00, 0x00, 0x3f, 0x3f },
+			{
+				&theme->button_iconify_active_unpressed,
+				theme->window_active_button_iconify_unpressed_image_color,
+			},
+			{
+				&theme->button_iconify_inactive_unpressed,
+				theme->window_inactive_button_iconify_unpressed_image_color,
+			},
+		},
+		{
+			"max",
+			{ 0x3f, 0x3f, 0x21, 0x21, 0x21, 0x3f },
+			{
+				&theme->button_maximize_active_unpressed,
+				theme->window_active_button_max_unpressed_image_color,
+			},
+			{
+				&theme->button_maximize_inactive_unpressed,
+				theme->window_inactive_button_max_unpressed_image_color,
+			},
+		},
+		{
+			"close",
+			{ 0x33, 0x3f, 0x1e, 0x1e, 0x3f, 0x33 },
+			{
+				&theme->button_close_active_unpressed,
+				theme->window_active_button_close_unpressed_image_color,
+			},
+			{
+				&theme->button_close_inactive_unpressed,
+				theme->window_inactive_button_close_unpressed_image_color,
+			},
+		},
+	};
+
+	char filename[4096] = {0};
+	for (size_t i = 0; i < ARRAY_SIZE(buttons); ++i) {
+		struct button *b = &buttons[i];
+
+		drop(b->active.buffer);
+		drop(b->inactive.buffer);
+
+		/* Try png icon first */
+		snprintf(filename, sizeof(filename), "%s-active.png", b->name);
+		button_png_load(filename, b->active.buffer);
+		snprintf(filename, sizeof(filename), "%s-inactive.png", b->name);
+		button_png_load(filename, b->inactive.buffer);
+
+#if HAVE_RSVG
+		/* Then try svg icon */
+		int size = theme->title_height - 2 * theme->padding_height;
+		if (!*b->active.buffer) {
+			snprintf(filename, sizeof(filename), "%s-active.svg", b->name);
+			button_svg_load(filename, b->active.buffer, size);
+		}
+		if (!*b->inactive.buffer) {
+			snprintf(filename, sizeof(filename), "%s-inactive.svg", b->name);
+			button_svg_load(filename, b->inactive.buffer, size);
+		}
+#endif
+
+		/* If there were no png/svg buttons, use xbm */
+		snprintf(filename, sizeof(filename), "%s.xbm", b->name);
+		if (!*b->active.buffer) {
+			button_xbm_load(filename, b->active.buffer,
+				b->fallback_button, b->active.rgba);
+		}
+		if (!*b->inactive.buffer) {
+			button_xbm_load(filename, b->inactive.buffer,
+				b->fallback_button, b->inactive.rgba);
+		}
+	}
+}
+
 static int
 hex_to_dec(char c)
 {
@@ -454,7 +573,21 @@ rounded_rect(struct rounded_corner_ctx *ctx)
 	cairo_set_operator(cairo, CAIRO_OPERATOR_CLEAR);
 	cairo_paint(cairo);
 
-	/* fill */
+	/*
+	 * Create outline path and fill. Illustration of top-left corner buffer:
+	 *
+	 *          _,,ooO"""""""""+
+	 *        ,oO"'   ^        |
+	 *      ,o"       |        |
+	 *     o"         |r       |
+	 *    o'          |        |
+	 *    O     r     v        |
+	 *    O<--------->+        |
+	 *    O                    |
+	 *    O                    |
+	 *    O                    |
+	 *    +--------------------+
+	 */
 	cairo_set_line_width(cairo, 0.0);
 	cairo_new_sub_path(cairo);
 	switch (ctx->corner) {
@@ -479,28 +612,102 @@ rounded_rect(struct rounded_corner_ctx *ctx)
 	cairo_fill_preserve(cairo);
 	cairo_stroke(cairo);
 
-	/* border */
-	cairo_set_line_cap(cairo, CAIRO_LINE_CAP_ROUND);
+	/*
+	 * Stroke horizontal and vertical borders, shown by Xs and Ys
+	 * respectively in the figure below:
+	 *
+	 *          _,,ooO"XXXXXXXXX
+	 *        ,oO"'            |
+	 *      ,o"                |
+	 *     o"                  |
+	 *    o'                   |
+	 *    O                    |
+	 *    Y                    |
+	 *    Y                    |
+	 *    Y                    |
+	 *    Y                    |
+	 *    Y--------------------+
+	 */
+	cairo_set_line_cap(cairo, CAIRO_LINE_CAP_BUTT);
 	set_cairo_color(cairo, ctx->border_color);
 	cairo_set_line_width(cairo, ctx->line_width);
 	double half_line_width = ctx->line_width / 2.0;
 	switch (ctx->corner) {
 	case LAB_CORNER_TOP_LEFT:
 		cairo_move_to(cairo, half_line_width, h);
-		cairo_line_to(cairo, half_line_width, r + half_line_width);
-		cairo_arc(cairo, r, r, r - half_line_width, 180 * deg, 270 * deg);
+		cairo_line_to(cairo, half_line_width, r);
+		cairo_move_to(cairo, r, half_line_width);
 		cairo_line_to(cairo, w, half_line_width);
 		break;
 	case LAB_CORNER_TOP_RIGHT:
 		cairo_move_to(cairo, 0, half_line_width);
 		cairo_line_to(cairo, w - r, half_line_width);
-		cairo_arc(cairo, w - r, r, r - half_line_width, -90 * deg, 0 * deg);
+		cairo_move_to(cairo, w - half_line_width, r);
 		cairo_line_to(cairo, w - half_line_width, h);
 		break;
 	default:
 		wlr_log(WLR_ERROR, "unknown corner type");
 	}
 	cairo_stroke(cairo);
+
+	/*
+	 * If radius==0 the borders stroked above go right up to (and including)
+	 * the corners, so there is not need to do any more.
+	 */
+	if (!r) {
+		goto out;
+	}
+
+	/*
+	 * Stroke the arc section of the border of the corner piece.
+	 *
+	 * Note: This figure is drawn at a more zoomed in scale compared with
+	 * those above.
+	 *
+	 *                 ,,ooooO""  ^
+	 *            ,ooo""'      |  |
+	 *         ,oOO"           |  | line-thickness
+	 *       ,OO"              |  |
+	 *     ,OO"         _,,ooO""  v
+	 *    ,O"         ,oO"'
+	 *   ,O'        ,o"
+	 *  ,O'        o"
+	 *  o'        o'
+	 *  O         O
+	 *  O---------O            +
+	 *       <----------------->
+	 *          radius
+	 *
+	 * We handle the edge-case where line-thickness > radius by merely
+	 * setting line-thickness = radius and in effect drawing a quadrant of a
+	 * circle. In this case the X and Y borders butt up against the arc and
+	 * overlap each other (as their line-thickessnes are greater than the
+	 * linethickness of the arc). As a result, there is no inner rounded
+	 * corners.
+	 *
+	 * So, in order to have inner rounded corners cornerRadius should be
+	 * greater than border.width.
+	 *
+	 * Also, see diagrams in https://github.com/labwc/labwc/pull/990
+	 */
+	double line_width = MIN(ctx->line_width, r);
+	cairo_set_line_width(cairo, line_width);
+	half_line_width = line_width / 2.0;
+	switch (ctx->corner) {
+	case LAB_CORNER_TOP_LEFT:
+		cairo_move_to(cairo, half_line_width, r);
+		cairo_arc(cairo, r, r, r - half_line_width, 180 * deg, 270 * deg);
+		break;
+	case LAB_CORNER_TOP_RIGHT:
+		cairo_move_to(cairo, w - r, half_line_width);
+		cairo_arc(cairo, w - r, r, r - half_line_width, -90 * deg, 0 * deg);
+		break;
+	default:
+		break;
+	}
+	cairo_stroke(cairo);
+
+out:
 	cairo_surface_flush(surf);
 
 	return buffer;
@@ -550,7 +757,7 @@ post_processing(struct theme *theme)
 		+ 2 * theme->osd_window_switcher_item_active_border_width;
 
 	if (rc.corner_radius >= theme->title_height) {
-		theme->title_height = rc.corner_radius + 1;
+		rc.corner_radius = theme->title_height - 1;
 	}
 
 	if (theme->menu_max_width < theme->menu_min_width) {
@@ -607,7 +814,7 @@ theme_init(struct theme *theme, const char *theme_name)
 
 	post_processing(theme);
 	create_corners(theme);
-	xbm_load(theme);
+	load_buttons(theme);
 }
 
 void
