@@ -7,6 +7,7 @@
 #include "labwc.h"
 #include "menu/menu.h"
 #include "regions.h"
+#include "resize_indicator.h"
 #include "ssd.h"
 #include "view.h"
 #include "window-rules.h"
@@ -18,6 +19,76 @@
 #define LAB_FALLBACK_WIDTH  640
 #define LAB_FALLBACK_HEIGHT 480
 
+static bool
+matches_criteria(struct view *view, enum lab_view_criteria criteria)
+{
+	if (!view_isfocusable(view)) {
+		return false;
+	}
+	if (criteria & LAB_VIEW_CRITERIA_CURRENT_WORKSPACE) {
+		/*
+		 * Always-on-top views are always on the current desktop and are
+		 * special in that they live in a different tree.
+		 */
+		struct server *server = view->server;
+		if (view->scene_tree->node.parent != server->workspace_current->tree
+				&& !view_is_always_on_top(view)) {
+			return false;
+		}
+	}
+	if (criteria & LAB_VIEW_CRITERIA_NO_ALWAYS_ON_TOP) {
+		if (view_is_always_on_top(view)) {
+			return false;
+		}
+	}
+	if (criteria & LAB_VIEW_CRITERIA_NO_SKIP_WINDOW_SWITCHER) {
+		if (window_rules_get_property(view, "skipWindowSwitcher") == LAB_PROP_TRUE) {
+			return false;
+		}
+	}
+	return true;
+}
+
+struct view *
+view_next(struct wl_list *head, struct view *view, enum lab_view_criteria criteria)
+{
+	assert(head);
+
+	struct wl_list *elm = view ? &view->link : head;
+
+	for (elm = elm->next; elm != head; elm = elm->next) {
+		view = wl_container_of(elm, view, link);
+		if (matches_criteria(view, criteria)) {
+			return view;
+		}
+	}
+	return NULL;
+}
+
+void
+view_array_append(struct server *server, struct wl_array *views,
+		enum lab_view_criteria criteria)
+{
+	struct view *view;
+	for_each_view(view, &server->views, criteria) {
+		struct view **entry = wl_array_add(views, sizeof(*entry));
+		if (!entry) {
+			wlr_log(WLR_ERROR, "wl_array_add(): out of memory");
+			continue;
+		}
+		*entry = view;
+	}
+}
+
+bool
+view_isfocusable(struct view *view)
+{
+	if (!view->surface) {
+		return false;
+	}
+	return (view->mapped || view->minimized);
+}
+
 /**
  * All view_apply_xxx_geometry() functions must *not* modify
  * any state besides repositioning or resizing the view.
@@ -25,16 +96,6 @@
  * They may be called repeatably during output layout changes.
  */
 
-enum view_edge {
-	VIEW_EDGE_INVALID = 0,
-
-	VIEW_EDGE_LEFT,
-	VIEW_EDGE_RIGHT,
-	VIEW_EDGE_UP,
-	VIEW_EDGE_DOWN,
-	VIEW_EDGE_CENTER,
-};
-
 static enum view_edge
 view_edge_invert(enum view_edge edge)
 {
@@ -192,6 +253,9 @@ view_moved(struct view *view)
 	if (view->toplevel.handle) {
 		foreign_toplevel_update_outputs(view);
 	}
+	if (rc.resize_indicator && view->server->grabbed_view == view) {
+		resize_indicator_update(view);
+	}
 }
 
 void
@@ -206,8 +270,9 @@ view_move_resize(struct view *view, struct wlr_box geo)
 void
 view_resize_relative(struct view *view, int left, int right, int top, int bottom)
 {
-	if (view->fullscreen || view->maximized)
+	if (view->fullscreen || view->maximized) {
 		return;
+	}
 	struct wlr_box newgeo = view->pending;
 	newgeo.x -= left;
 	newgeo.width += left + right;
@@ -217,6 +282,20 @@ view_resize_relative(struct view *view, int left, int right, int top, int bottom
 	view_set_untiled(view);
 }
 
+void
+view_move_relative(struct view *view, int x, int y)
+{
+	if (view->fullscreen) {
+		return;
+	}
+	view_maximize(view, false, /*store_natural_geometry*/ false);
+	if (view_is_tiled(view)) {
+		view_set_untiled(view);
+		view_restore_to(view, view->natural_geometry);
+	}
+	view_move(view, view->pending.x + x, view->pending.y + y);
+}
+
 void
 view_adjust_size(struct view *view, int *w, int *h)
 {
@@ -233,8 +312,8 @@ view_adjust_size(struct view *view, int *w, int *h)
 	*h = MAX(*h, LAB_MIN_VIEW_HEIGHT);
 }
 
-void
-view_minimize(struct view *view, bool minimized)
+static void
+_minimize(struct view *view, bool minimized)
 {
 	assert(view);
 	if (view->minimized == minimized) {
@@ -267,6 +346,40 @@ view_minimize(struct view *view, bool minimized)
 	}
 }
 
+static void
+minimize_sub_views(struct view *view, bool minimized)
+{
+	struct view **child;
+	struct wl_array children;
+
+	wl_array_init(&children);
+	view_append_children(view, &children);
+	wl_array_for_each(child, &children) {
+		_minimize(*child, minimized);
+		minimize_sub_views(*child, minimized);
+	}
+	wl_array_release(&children);
+}
+
+/*
+ * Minimize the whole view-hierarchy from top to bottom regardless of which one
+ * in the hierarchy requested the minimize. For example, if an 'About' or
+ * 'Open File' dialog is minimized, its toplevel is minimized also. And vice
+ * versa.
+ */
+void
+view_minimize(struct view *view, bool minimized)
+{
+	/*
+	 * Minimize the root window first because some xwayland clients send a
+	 * request-unmap to sub-windows at this point (for example gimp and its
+	 * 'open file' dialog), so it saves trying to unmap them twice
+	 */
+	struct view *root = view_get_root(view);
+	_minimize(root, minimized);
+	minimize_sub_views(root, minimized);
+}
+
 static bool
 view_compute_centered_position(struct view *view, const struct wlr_box *ref,
 		int w, int h, int *x, int *y)
@@ -535,6 +648,13 @@ set_maximized(struct view *view, bool maximized)
 			view->toplevel.handle, maximized);
 	}
 	view->maximized = maximized;
+
+	/*
+	 * Ensure that follow-up actions like SnapToEdge / SnapToRegion
+	 * use up-to-date SSD margin information. Otherwise we will end
+	 * up using an outdated ssd->margin to calculate offsets.
+	 */
+	ssd_update_margin(view->ssd);
 }
 
 /*
@@ -619,6 +739,19 @@ void
 view_toggle_decorations(struct view *view)
 {
 	assert(view);
+	if (rc.ssd_keep_border && view->ssd_enabled && view->ssd
+			&& !view->ssd_titlebar_hidden) {
+		/*
+		 * ssd_titlebar_hidden has to be set before calling
+		 * ssd_titlebar_hide() to make ssd_thickness() happy.
+		 */
+		view->ssd_titlebar_hidden = true;
+		ssd_titlebar_hide(view->ssd);
+		if (!view_is_floating(view)) {
+			view_apply_special_geometry(view);
+		}
+		return;
+	}
 	view_set_decorations(view, !view->ssd_enabled);
 }
 
@@ -709,6 +842,7 @@ view_set_decorations(struct view *view, bool decorations)
 			decorate(view);
 		} else {
 			undecorate(view);
+			view->ssd_titlebar_hidden = false;
 		}
 		if (!view_is_floating(view)) {
 			view_apply_special_geometry(view);
@@ -843,7 +977,7 @@ view_on_output_destroy(struct view *view)
 }
 
 void
-view_move_to_edge(struct view *view, const char *direction)
+view_move_to_edge(struct view *view, enum view_edge edge)
 {
 	assert(view);
 	struct output *output = view->output;
@@ -851,10 +985,6 @@ view_move_to_edge(struct view *view, const char *direction)
 		wlr_log(WLR_ERROR, "view has no output, not moving to edge");
 		return;
 	}
-	if (!direction) {
-		wlr_log(WLR_ERROR, "invalid edge, not moving view");
-		return;
-	}
 
 	struct border margin = ssd_get_margin(view->ssd);
 	struct wlr_box usable = output_usable_area_in_layout_coords(output);
@@ -868,28 +998,32 @@ view_move_to_edge(struct view *view, const char *direction)
 	}
 
 	int x = 0, y = 0;
-	if (!strcasecmp(direction, "left")) {
+	switch (edge) {
+	case VIEW_EDGE_LEFT:
 		x = usable.x + margin.left + rc.gap;
 		y = view->pending.y;
-	} else if (!strcasecmp(direction, "up")) {
+		break;
+	case VIEW_EDGE_UP:
 		x = view->pending.x;
 		y = usable.y + margin.top + rc.gap;
-	} else if (!strcasecmp(direction, "right")) {
+		break;
+	case VIEW_EDGE_RIGHT:
 		x = usable.x + usable.width - view->pending.width
 			- margin.right - rc.gap;
 		y = view->pending.y;
-	} else if (!strcasecmp(direction, "down")) {
+		break;
+	case VIEW_EDGE_DOWN:
 		x = view->pending.x;
 		y = usable.y + usable.height - view->pending.height
 			- margin.bottom - rc.gap;
-	} else {
-		wlr_log(WLR_ERROR, "invalid edge, not moving view");
+		break;
+	default:
 		return;
 	}
 	view_move(view, x, y);
 }
 
-static enum view_edge
+enum view_edge
 view_edge_parse(const char *direction)
 {
 	if (!direction) {
@@ -911,8 +1045,7 @@ view_edge_parse(const char *direction)
 }
 
 void
-view_snap_to_edge(struct view *view, const char *direction,
-		bool store_natural_geometry)
+view_snap_to_edge(struct view *view, enum view_edge edge, bool store_natural_geometry)
 {
 	assert(view);
 	if (view->fullscreen) {
@@ -923,11 +1056,6 @@ view_snap_to_edge(struct view *view, const char *direction,
 		wlr_log(WLR_ERROR, "view has no output, not snapping to edge");
 		return;
 	}
-	enum view_edge edge = view_edge_parse(direction);
-	if (edge == VIEW_EDGE_INVALID) {
-		wlr_log(WLR_ERROR, "invalid edge, not snapping view");
-		return;
-	}
 
 	if (view->tiled == edge && !view->maximized) {
 		/* We are already tiled for this edge and thus should switch outputs */
@@ -1043,6 +1171,25 @@ view_move_to_back(struct view *view)
 	}
 }
 
+struct view *
+view_get_root(struct view *view)
+{
+	assert(view);
+	if (view->impl->get_root) {
+		return view->impl->get_root(view);
+	}
+	return view;
+}
+
+void
+view_append_children(struct view *view, struct wl_array *children)
+{
+	assert(view);
+	if (view->impl->append_children) {
+		view->impl->append_children(view, children);
+	}
+}
+
 const char *
 view_get_string_prop(struct view *view, const char *prop)
 {
@@ -1198,3 +1345,4 @@ view_destroy(struct view *view)
 		cursor_update_focus(server);
 	}
 }
+
