@@ -320,6 +320,72 @@ xdg_toplevel_view_minimize(struct view *view, bool minimized)
 	/* noop */
 }
 
+static struct wlr_xdg_toplevel *
+top_parent_of(struct view *view)
+{
+	struct wlr_xdg_toplevel *toplevel = xdg_toplevel_from_view(view);
+	while (toplevel->parent) {
+		toplevel = toplevel->parent;
+	}
+	return toplevel;
+}
+
+/* Return the most senior parent (=root) view */
+static struct view *
+xdg_toplevel_view_get_root(struct view *view)
+{
+	struct wlr_xdg_toplevel *root = top_parent_of(view);
+	struct wlr_xdg_surface *surface = (struct wlr_xdg_surface *)root->base;
+	return (struct view *)surface->data;
+}
+
+/*
+ * In the view_move_to_{front,back} functions, a modal dialog is always shown
+ * above its parent window, and the two always move together, so other window
+ * cannot come between them.
+ * This is consistent with GTK3/Qt5 applications on mutter and openbox.
+ */
+static void
+xdg_toplevel_view_move_to_front(struct view *view)
+{
+	struct view *root = xdg_toplevel_view_get_root(view);
+	view_impl_move_to_front(root);
+	view_impl_move_sub_views(root, LAB_TO_FRONT);
+}
+
+static void
+xdg_toplevel_view_move_to_back(struct view *view)
+{
+	struct view *root = xdg_toplevel_view_get_root(view);
+	view_impl_move_sub_views(root, LAB_TO_BACK);
+	view_impl_move_to_back(root);
+}
+
+static void
+xdg_toplevel_view_append_children(struct view *self, struct wl_array *children)
+{
+	struct wlr_xdg_toplevel *toplevel = xdg_toplevel_from_view(self);
+	struct view *view;
+
+	wl_list_for_each_reverse(view, &self->server->views, link)
+	{
+		if (view == self) {
+			continue;
+		}
+		if (view->type != LAB_XDG_SHELL_VIEW) {
+			continue;
+		}
+		if (!view->mapped && !view->minimized) {
+			continue;
+		}
+		if (top_parent_of(view) != toplevel) {
+			continue;
+		}
+		struct view **child = wl_array_add(children, sizeof(*child));
+		*child = view;
+	}
+}
+
 static void
 xdg_toplevel_view_set_activated(struct view *view, bool activated)
 {
@@ -373,7 +439,19 @@ position_xdg_toplevel_view(struct view *view)
 static const char *
 xdg_toplevel_view_get_string_prop(struct view *view, const char *prop)
 {
-	struct wlr_xdg_toplevel *xdg_toplevel = xdg_toplevel_from_view(view);
+	struct xdg_toplevel_view *xdg_view = xdg_toplevel_view_from_view(view);
+	struct wlr_xdg_toplevel *xdg_toplevel = xdg_view->xdg_surface
+		? xdg_view->xdg_surface->toplevel
+		: NULL;
+	if (!xdg_toplevel) {
+		/*
+		 * This may happen due to a matchOnce rule when
+		 * a view is destroyed while A-Tab is open. See
+		 * https://github.com/labwc/labwc/issues/1082#issuecomment-1716137180
+		 */
+		return "";
+	}
+
 	if (!strcmp(prop, "title")) {
 		return xdg_toplevel->title;
 	}
@@ -383,6 +461,22 @@ xdg_toplevel_view_get_string_prop(struct view *view, const char *prop)
 	return "";
 }
 
+static void
+init_foreign_toplevel(struct view *view)
+{
+	foreign_toplevel_handle_create(view);
+	struct wlr_xdg_toplevel *toplevel = xdg_toplevel_from_view(view);
+	if (!toplevel->parent) {
+		return;
+	}
+	struct wlr_xdg_surface *surface = toplevel->parent->base;
+	struct view *parent = surface->data;
+	if (!parent->toplevel.handle) {
+		return;
+	}
+	wlr_foreign_toplevel_handle_v1_set_parent(view->toplevel.handle, parent->toplevel.handle);
+}
+
 static void
 xdg_toplevel_view_map(struct view *view)
 {
@@ -399,7 +493,9 @@ xdg_toplevel_view_map(struct view *view)
 	if (!view->been_mapped) {
 		struct wlr_xdg_toplevel_requested *requested =
 			&xdg_toplevel_from_view(view)->requested;
-		foreign_toplevel_handle_create(view);
+
+		init_foreign_toplevel(view);
+
 		view_set_decorations(view, has_ssd(view));
 
 		/*
@@ -472,8 +568,10 @@ static const struct view_impl xdg_toplevel_view_impl = {
 	.unmap = xdg_toplevel_view_unmap,
 	.maximize = xdg_toplevel_view_maximize,
 	.minimize = xdg_toplevel_view_minimize,
-	.move_to_front = view_impl_move_to_front,
-	.move_to_back = view_impl_move_to_back,
+	.move_to_front = xdg_toplevel_view_move_to_front,
+	.move_to_back = xdg_toplevel_view_move_to_back,
+	.get_root = xdg_toplevel_view_get_root,
+	.append_children = xdg_toplevel_view_append_children,
 };
 
 void
@@ -504,6 +602,11 @@ xdg_activation_handle_request(struct wl_listener *listener, void *data)
 	 * for the seat / serial being correct and then allow the request.
 	 */
 
+	if (window_rules_get_property(view, "ignoreFocusRequest") == LAB_PROP_TRUE) {
+		wlr_log(WLR_INFO, "Ignoring focus request due to window rule configuration");
+		return;
+	}
+
 	/*
 	 * TODO: This is the exact same code as used in foreign.c.
 	 *       Refactor it into a public helper function somewhere.
