@@ -16,8 +16,9 @@
 static int
 round_to_increment(int val, int base, int inc)
 {
-	if (base < 0 || inc <= 0)
+	if (base < 0 || inc <= 0) {
 		return val;
+	}
 	return base + (val - base + inc / 2) / inc * inc;
 }
 
@@ -49,6 +50,21 @@ xwayland_apply_size_hints(struct view *view, int *w, int *h)
 	return false;
 }
 
+static void
+xwayland_view_fill_size_hints(struct view *view, struct wlr_box *box)
+{
+	if (view->type == LAB_XWAYLAND_VIEW) {
+		xcb_size_hints_t *hints = xwayland_surface_from_view(view)->size_hints;
+		if (hints) {
+			box->width = hints->width_inc;
+			box->height = hints->height_inc;
+			return;
+		}
+	}
+	box->width = 0;
+	box->height = 0;
+}
+
 static struct wlr_xwayland_surface *
 top_parent_of(struct view *view)
 {
@@ -201,7 +217,6 @@ handle_destroy(struct wl_listener *listener, void *data)
 {
 	struct view *view = wl_container_of(listener, view, destroy);
 	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
-	assert(data && data == xwayland_view->xwayland_surface);
 	assert(xwayland_view->xwayland_surface->data == view);
 
 	if (view->surface) {
@@ -270,6 +285,12 @@ handle_request_activate(struct wl_listener *listener, void *data)
 	struct xwayland_view *xwayland_view =
 		wl_container_of(listener, xwayland_view, request_activate);
 	struct view *view = &xwayland_view->base;
+
+	if (window_rules_get_property(view, "ignoreFocusRequest") == LAB_PROP_TRUE) {
+		wlr_log(WLR_INFO, "Ignoring focus request due to window rule configuration");
+		return;
+	}
+
 	desktop_focus_and_activate_view(&view->server->seat, view);
 	view_move_to_front(view);
 }
@@ -334,8 +355,17 @@ xwayland_view_close(struct view *view)
 static const char *
 xwayland_view_get_string_prop(struct view *view, const char *prop)
 {
-	struct wlr_xwayland_surface *xwayland_surface =
-		xwayland_surface_from_view(view);
+	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
+	struct wlr_xwayland_surface *xwayland_surface = xwayland_view->xwayland_surface;
+	if (!xwayland_surface) {
+		/*
+		 * This may happen due to a matchOnce rule when
+		 * a view is destroyed while A-Tab is open. See
+		 * https://github.com/labwc/labwc/issues/1082#issuecomment-1716137180
+		 */
+		return "";
+	}
+
 	if (!strcmp(prop, "title")) {
 		return xwayland_surface->title;
 	}
@@ -355,7 +385,6 @@ handle_set_decorations(struct wl_listener *listener, void *data)
 	struct xwayland_view *xwayland_view =
 		wl_container_of(listener, xwayland_view, set_decorations);
 	struct view *view = &xwayland_view->base;
-	assert(data && data == xwayland_view->xwayland_surface);
 
 	view_set_decorations(view, want_deco(xwayland_view->xwayland_surface));
 }
@@ -366,8 +395,7 @@ handle_override_redirect(struct wl_listener *listener, void *data)
 	struct xwayland_view *xwayland_view =
 		wl_container_of(listener, xwayland_view, override_redirect);
 	struct view *view = &xwayland_view->base;
-	struct wlr_xwayland_surface *xsurface = data;
-	assert(xsurface && xsurface == xwayland_view->xwayland_surface);
+	struct wlr_xwayland_surface *xsurface = xwayland_view->xwayland_surface;
 
 	struct server *server = view->server;
 	bool mapped = xsurface->mapped;
@@ -410,17 +438,42 @@ top_left_edge_boundary_check(struct view *view)
 	view->impl->configure(view, view->current);
 }
 
+static void
+init_foreign_toplevel(struct view *view)
+{
+	foreign_toplevel_handle_create(view);
+
+	struct wlr_xwayland_surface *surface = xwayland_surface_from_view(view);
+	if (!surface->parent) {
+		return;
+	}
+	struct view *parent = (struct view *)surface->parent->data;
+	if (!parent->toplevel.handle) {
+		return;
+	}
+	wlr_foreign_toplevel_handle_v1_set_parent(view->toplevel.handle, parent->toplevel.handle);
+}
+
 static void
 xwayland_view_map(struct view *view)
 {
+	struct wlr_xwayland_surface *xwayland_surface = xwayland_surface_from_view(view);
 	if (view->mapped) {
 		return;
 	}
+	if (!xwayland_surface->surface) {
+		/*
+		 * We may get here if a user minimizes an xwayland dialog at the
+		 * same time as the client requests unmap, which xwayland
+		 * clients sometimes do without actually requesting destroy
+		 * even if they don't intend to use that view/surface anymore
+		 */
+		wlr_log(WLR_DEBUG, "Cannot map view without wlr_surface");
+		return;
+	}
 	view->mapped = true;
 	ensure_initial_geometry_and_output(view);
 	wlr_scene_node_set_enabled(&view->scene_tree->node, true);
-	struct wlr_xwayland_surface *xwayland_surface =
-		xwayland_surface_from_view(view);
 	if (!view->fullscreen && xwayland_surface->fullscreen) {
 		view_set_fullscreen(view, true);
 	}
@@ -447,7 +500,7 @@ xwayland_view_map(struct view *view)
 	}
 
 	if (!view->toplevel.handle) {
-		foreign_toplevel_handle_create(view);
+		init_foreign_toplevel(view);
 	}
 
 	if (!view->been_mapped) {
@@ -483,7 +536,7 @@ static void
 xwayland_view_unmap(struct view *view, bool client_request)
 {
 	if (!view->mapped) {
-		return;
+		goto out;
 	}
 	view->mapped = false;
 	wl_list_remove(&view->commit.link);
@@ -495,6 +548,7 @@ xwayland_view_unmap(struct view *view, bool client_request)
 	 * than just minimized), destroy the foreign toplevel handle so
 	 * the unmapped view doesn't show up in panels and the like.
 	 */
+out:
 	if (client_request && view->toplevel.handle) {
 		wlr_foreign_toplevel_handle_v1_destroy(view->toplevel.handle);
 		view->toplevel.handle = NULL;
@@ -515,58 +569,59 @@ xwayland_view_minimize(struct view *view, bool minimized)
 		minimized);
 }
 
-enum z_direction {
-	LAB_TO_FRONT,
-	LAB_TO_BACK,
-};
-
-static void
-move_sub_views(struct view *parent, enum z_direction z_direction)
+static struct view *
+xwayland_view_get_root(struct view *view)
 {
-	assert(parent);
-
-	if (parent->type != LAB_XWAYLAND_VIEW) {
-		return;
-	}
-
-	struct wlr_xwayland_surface *parent_xwayland_surface =
-		xwayland_surface_from_view(parent);
-	struct view *view, *next;
-	wl_list_for_each_reverse_safe(view, next, &parent->server->views, link)
-	{
-		/* need to stop here, otherwise loops keeps going forever */
-		if (view == parent) {
-			break;
-		}
-		if (view->type != LAB_XWAYLAND_VIEW) {
-			continue;
-		}
-		if (!view->mapped && !view->minimized) {
-			continue;
-		}
-		if (top_parent_of(view) != parent_xwayland_surface) {
-			continue;
-		}
-		if (z_direction == LAB_TO_FRONT) {
-			view_impl_move_to_front(view);
-		} else if (z_direction == LAB_TO_BACK) {
-			view_impl_move_to_back(view);
-		}
-	}
+	struct wlr_xwayland_surface *root = top_parent_of(view);
+	return (struct view *)root->data;
 }
 
 static void
 xwayland_view_move_to_front(struct view *view)
 {
-	view_impl_move_to_front(view);
-	move_sub_views(view, LAB_TO_FRONT);
+	struct view *root = xwayland_view_get_root(view);
+	view_impl_move_to_front(root);
+	view_impl_move_sub_views(root, LAB_TO_FRONT);
 }
 
 static void
 xwayland_view_move_to_back(struct view *view)
 {
-	view_impl_move_to_back(view);
-	move_sub_views(view, LAB_TO_BACK);
+	struct view *root = xwayland_view_get_root(view);
+	view_impl_move_sub_views(root, LAB_TO_BACK);
+	view_impl_move_to_back(root);
+}
+
+static void
+xwayland_view_append_children(struct view *self, struct wl_array *children)
+{
+	struct wlr_xwayland_surface *surface = xwayland_surface_from_view(self);
+	struct view *view;
+
+	wl_list_for_each_reverse(view, &self->server->views, link)
+	{
+		if (view == self) {
+			continue;
+		}
+		if (view->type != LAB_XWAYLAND_VIEW) {
+			continue;
+		}
+		/*
+		 * This happens when a view has never been mapped or when a
+		 * client has requested a `handle_unmap`.
+		 */
+		if (!view->surface) {
+			continue;
+		}
+		if (!view->mapped && !view->minimized) {
+			continue;
+		}
+		if (top_parent_of(view) != surface) {
+			continue;
+		}
+		struct view **child = wl_array_add(children, sizeof(*child));
+		*child = view;
+	}
 }
 
 static void
@@ -612,6 +667,9 @@ static const struct view_impl xwayland_view_impl = {
 	.minimize = xwayland_view_minimize,
 	.move_to_front = xwayland_view_move_to_front,
 	.move_to_back = xwayland_view_move_to_back,
+	.get_root = xwayland_view_get_root,
+	.append_children = xwayland_view_append_children,
+	.fill_size_hints = xwayland_view_fill_size_hints,
 };
 
 void
@@ -748,6 +806,11 @@ xwayland_server_init(struct server *server, struct wlr_compositor *compositor)
 void
 xwayland_server_finish(struct server *server)
 {
-	wlr_xwayland_destroy(server->xwayland);
+	struct wlr_xwayland *xwayland = server->xwayland;
+	/*
+	 * Reset server->xwayland to NULL first to prevent callbacks (like
+	 * server_global_filter) from accessing it as it is destroyed
+	 */
 	server->xwayland = NULL;
+	wlr_xwayland_destroy(xwayland);
 }
